(* $Id: main.ml,v 1.30 2012-03-05 10:52:52 deraugla Exp $ *)

open Printf;
open Ast;

value version = "1.3";

value schedule_loop (quiet, colour) ts task_pri = do {
  let tc = ts.constraints in
  let accu_itask = task_pri.(0) in
  let accu_task = Taskalg.of_task_def accu_itask ts.tasks_def.(accu_itask) in
  if quiet then ()
  else do {
    printf "t%d =" (accu_itask + 1);
    printf " %s\n" (Taskalg.to_string colour accu_task);
  };
  loop [accu_itask] 1 accu_task where rec loop stl i accu_task =
    if i >= Array.length ts.tasks_def then Left accu_task
    else do {
      let itask2 = task_pri.(i) in
      let stask2 = Taskalg.of_task_def itask2 ts.tasks_def.(itask2) in
      let stl = [itask2 :: stl] in
      if quiet then ()
      else do {
        let sl = List.rev_map (fun i -> sprintf "t%d" (i + 1)) stl in
        let s = String.concat " ⊕ " sl in
        printf "%s =" s;
        flush stdout;
      };
      match Taskalg.plus tc accu_task stask2 with
      [ Left accu_task -> do {
          if quiet then ()
          else do {
            printf " %s\n" (Taskalg.to_string colour accu_task);
            flush stdout;
          };
          loop stl (i+1) accu_task
        }
      | Right msg -> do {
          if quiet then ()
          else do {
            printf "\n";
            flush stdout;
          };
          Right msg
        } ]
    }
};

value schedule_edf quiet ts = do {
  let tl =
    loop [] 0 where rec loop rev_tl itask =
      if itask = Array.length ts.tasks_def then List.rev rev_tl
      else do {
        let t = Taskalg.of_task_def itask ts.tasks_def.(itask) in
        let rev_tl = [t :: rev_tl] in
        loop rev_tl (itask + 1)
      }
  in
  Taskalg.edf ts.constraints tl
};

value array_rev a =
  let len = Array.length a in
  Array.init len (fun i -> a.(len - i - 1))
;

value do_schedule (quiet, algo, colour) df = do {
  let ts = Sparse.make_ts df in
  let ts =
    match algo with
    [ Some a -> {(ts) with algorithm = a}
    | None -> ts ]
  in
  let tc = ts.constraints in
  if quiet then ()
  else do {
    for i = 0 to Array.length ts.tasks_def - 1 do {
      let td = ts.tasks_def.(i) in
      printf "t%d: rel %d dur %d" (i + 1) td.release td.duration;
      match td.sub_duration with
      [ [] -> ()
      | _ -> printf "..." ];
      if td.deadline <> td.period then printf " dea %d" td.deadline else ();
      printf " per %d" td.period;
      printf "\n";
      flush stdout;
      let st = Taskalg.of_task_def i td in
      printf "t%d: " (i + 1);
      printf "%s" (Taskalg.to_string colour st);
      printf "\n";
      flush stdout;
    }
  };
  let task_pri = do {
(*
    let a = Array.init (Array.length ts.tasks_def) (fun i -> i) in
    Array.sort (Sparse.compare_priorities ts) a;
    a
*)
    let len = Array.length ts.tasks_def in
    Array.init len (fun i -> len - 1 - i)
(**)
  }
  in
  if quiet then ()
  else do {
    if ts.algorithm <> AL_edf && tc.graph_precedence <> [] then do {
      printf "precedence:";
      List.iter
        (fun tl ->
           let s =
             String.concat "<" (List.map (fun i -> sprintf "t%d" (i + 1)) tl)
           in
           printf " %s" s)
        tc.graph_precedence;
      printf "\n";
      flush stdout;
    }
    else ();
    printf "algorithm: ";
    match ts.algorithm with
    [ AL_deadline_monotonic -> printf "DM"
    | AL_rate_monotonic -> printf "RM"
    | AL_edf -> printf "EDF"
    | AL_audsley _ | AL_user _ -> failwith "internal error: bad algo" ];
    printf "\n";
    if ts.algorithm <> AL_edf then do {
      printf "\ntasks priorities:";
      for i = 0 to Array.length ts.tasks_def - 1 do {
        printf " t%d" (task_pri.(i) + 1);
      };
      printf "\n";
    }
    else ();
    printf "\n";
    flush stdout;
  };
  if Array.length task_pri = 0 then do {
    if quiet then ()
    else do {
      printf "no task\n";
      flush stdout;
    };
    0
  }
  else do {
    let accu_task_opt =
      if ts.algorithm = AL_edf then schedule_edf quiet ts
      else schedule_loop (quiet, colour) ts task_pri
    in
    if quiet then ()
    else do {
      match accu_task_opt with
      [ Left accu_task -> do {
          printf "r = %s\n" (Taskalg.to_string colour accu_task)
        }
      | Right msg -> printf "not schedulable: %s\n" msg ];
      flush stdout;
    };
    match accu_task_opt with
    [ Left ts -> do {
        printf "FRT";
        let a = Taskalg.first_response_time ts in
        Array.sort compare a;
        Array.iter (fun (_, wrt) -> printf " %d" wrt) a;
        printf "\n";
        flush stdout;
        printf "LRT";
        let a = Taskalg.last_response_time ts in
(*
        Array.sort compare a;
*)
        let a = array_rev a in
(**)
        Array.iter (fun (_, wrt) -> printf " %d" wrt) a;
        printf "\n";
        flush stdout;
        printf "WRT";
        let a = Taskalg.worst_response_time ts in
        Array.sort compare a;
        Array.iter (fun (_, wrt) -> printf " %d" wrt) a;
        printf "\n";
        flush stdout;
      }
    | Right _ -> () ];
    match accu_task_opt with
    [ Left _ -> 0
    | Right _ -> 1 ]
  }
};

value taskadd (quiet, algo, colour) fname strm close = do {
  match
    try Left (Sparse.parse_and_check_dameid strm) with
    [ Ploc.Exc loc e -> Right (loc, e) ]
  with
  [ Left dl -> do {
      let r =
        List.fold_left
          (fun r (_, d) -> max r (do_schedule (quiet, algo, colour) d))
          0 dl
      in
      exit r
    }
  | Right (loc, e) -> do {
      if fname = "" then do {
        let line = Ploc.line_nb loc in
        let bp = Ploc.first_pos loc in
        let ep = Ploc.last_pos loc in
        eprintf "Line %d, characters %d-%d:\n" line bp ep;
      }
      else do {
        let (fname, line, bp, ep) = Ploc.from_file fname loc in
        eprintf "File \"%s\", line %d, characters %d-%d:\n" fname line bp ep;
      };
      match e with
      [ Stream.Error s -> eprintf "Parse error: %s\n" s
      | Plexing.Error s -> eprintf "Lexing error: %s\n" s
      | _ -> raise e ];
      flush stderr;
      exit 2;
    } ]
};

value arg_algo = ref None;
value arg_colour = ref False;
value arg_quiet = ref False;
value arg_string = ref False;
value arg_version = ref False;
value arg_value = ref "";

value set_algo arg_algo s = do {
  arg_algo.val :=
    match String.lowercase s with
    [ "dm" -> Some AL_deadline_monotonic
    | "rm" -> Some AL_rate_monotonic
    | "edf" -> Some AL_edf
    | _ -> raise (Arg.Bad (sprintf "Bad algorithm: \"%s\"" s)) ];
};

value spec_list =
  Arg.align
    [("-a", Arg.String (set_algo arg_algo),
     "<algo> Set algorithm (dm, rm, edf)");
     ("-c", Arg.Set arg_colour,
     " Colour display (recommended if your terminal accepts it)");
     ("-l", Arg.Set Taskalg.long_display,
      " Long: complete display of result even if very long");
     ("-q", Arg.Set arg_quiet,
      " Quiet: no output; exit 0 if schedulable, 1 if not");
     ("-s", Arg.Set arg_string,
      " Argument is string instead of input file name");
     ("-v", Arg.Set arg_version,
      " Display current version and exit")]
;
value anon_fun s =
  arg_value.val :=
    if not arg_string.val || arg_value.val = "" then s
    else arg_value.val ^ " " ^ s
;
value usage = "Usage: " ^ Sys.argv.(0) ^ " [options] filename-or-string";

value main () = do {
  Arg.parse spec_list anon_fun usage;
  if arg_version.val then do {
    printf "Taskadd version %s\n" version;
    flush stdout;
    exit 0;
  }
  else if arg_value.val = "" then do {
    eprintf "Error: use option -help for usage\n";
    flush stderr;
    exit 2;
  }
  else do {
    let (strm, fname, close) =
      if arg_string.val then do {
        let s = Plexing.eval_string Ploc.dummy arg_value.val in
        let s = s ^ "\n" in
        (Stream.of_string s, "", fun () -> ())
      }
      else do {
        let ic = open_in arg_value.val in
        let close () = close_in ic in
        (Stream.of_channel ic, arg_value.val, close)
      }
    in
    taskadd (arg_quiet.val, arg_algo.val, arg_colour.val) fname strm close
  }
};

main ();
